/*
 * Copyright (C) 2021 Mathieu Desnoyers <mathieu.desnoyers@efficios.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-only
 */

#define _LGPL_SOURCE
#include "hashtable.h"
#include "utils.h"

unsigned long lttng_ht_seed;
