/*
 * Copyright (C) 2016 Mathieu Desnoyers <mathieu.desnoyers@efficios.com>
 *
 * SPDX-License-Identifier: GPL-2.0-only
 *
 */

#ifndef _LIBBAR_H
#define _LIBBAR_H

int bar(void);

#endif  /* _LIBBAR_H */
