/*
 * Copyright (C) 2018 Francis Deslauriers <francis.deslauriers@efficios.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-only
 *
 */

#ifndef CALLSITES_H
#define CALLSITES_H
void call_tracepoint(void);
#endif /* CALLSITES_H */
